#include <stdlib.h>
#include <limits.h>
#include <string.h>

#include <iostream.h>
#include <strstream.h>

#include "colorbartrue24.h"
#include "util.h"

// Tk Canvas Widget Function Declarations

int ColorbarTrueColor24CreateProc(Tcl_Interp*, Tk_Canvas, Tk_Item*, int, 
				  char**);

// Colorbar Specs

static Tk_CustomOption tagsOption = {
  Tk_CanvasTagsParseProc, Tk_CanvasTagsPrintProc, NULL
};

class ColorbarTrueColor24Options : public WidgetOptions {
public:
  int colors;
};

static Tk_ConfigSpec colorbarTrueColor24Specs[] = {

  {TK_CONFIG_STRING, "-command", NULL, NULL, "colorbar",
   Tk_Offset(ColorbarTrueColor24Options, cmdName), 
   TK_CONFIG_OPTION_SPECIFIED, NULL},
  {TK_CONFIG_INT, "-x", NULL, NULL, "1",
   Tk_Offset(ColorbarTrueColor24Options, x), TK_CONFIG_OPTION_SPECIFIED, NULL},
  {TK_CONFIG_INT, "-y", NULL, NULL, "1",
   Tk_Offset(ColorbarTrueColor24Options, y), TK_CONFIG_OPTION_SPECIFIED, NULL},
  {TK_CONFIG_INT, "-width", NULL, NULL, "512",
   Tk_Offset(ColorbarTrueColor24Options, width), 
   TK_CONFIG_OPTION_SPECIFIED, NULL},
  {TK_CONFIG_INT, "-height", NULL, NULL, "22",
   Tk_Offset(ColorbarTrueColor24Options, height), 
   TK_CONFIG_OPTION_SPECIFIED, NULL},
  {TK_CONFIG_ANCHOR, "-anchor", NULL, NULL, "nw",
   Tk_Offset(ColorbarTrueColor24Options, anchor), 0, NULL},
  {TK_CONFIG_CUSTOM, "-tags", NULL, NULL, NULL,
   0, TK_CONFIG_NULL_OK, &tagsOption},

  {TK_CONFIG_INT, "-colors", NULL, NULL, "1024",
   Tk_Offset(ColorbarTrueColor24Options, colors), 0, NULL},

  {TK_CONFIG_END, NULL, NULL, NULL, NULL, 0, 0, NULL},
};

// Tk Static Structure

static Tk_ItemType colorbarTrueColor24Type = {
  "colorbartruecolor24",        // name
  sizeof(ColorbarTrueColor24Options), // size
  ColorbarTrueColor24CreateProc, // configProc
  colorbarTrueColor24Specs,     // configSpecs
  WidgetConfigProc,             // configProc
  WidgetCoordProc,              // coordProc
  WidgetDeleteProc,             // deleteProc
  WidgetDisplayProc,            // displayProc
  0,                            // alwaysRedraw
  WidgetPointProc,              // pointProc
  WidgetAreaProc,               // areaProc
  WidgetPostscriptProc,         // postscriptProc
  WidgetScaleProc,              // scaleProc
  WidgetTranslateProc,          // translateProc
  (Tk_ItemIndexProc*)NULL,      // indexProc
  (Tk_ItemCursorProc*)NULL,     // icursorProc
  (Tk_ItemSelectionProc*)NULL,  // selectionProc
  (Tk_ItemInsertProc*)NULL,     // insertProc
  (Tk_ItemDCharsProc*)NULL,     // dCharsProc
  (Tk_ItemType*)NULL            // nextPtr
};

// Non-Member Functions

int ColorbarTrueColor24_Init(Tcl_Interp* interp)
{
  Tk_CreateItemType(&colorbarTrueColor24Type);
  return TCL_OK;
}

int ColorbarTrueColor24CreateProc(Tcl_Interp* interp, Tk_Canvas canvas, 
				  Tk_Item* item, int argc, char** argv)
{
  ColorbarTrueColor24* colorbar = new ColorbarTrueColor24(interp,canvas,item);

  // and set default configuration

  if (colorbar->configure(argc, argv, 0) != TCL_OK) {
    delete colorbar;
    Tcl_AppendResult(interp, " error occured while creating colorbar.", NULL);
    return TCL_ERROR;
  }

  return TCL_OK;
}

// ColorbarTrueColor24

ColorbarTrueColor24::ColorbarTrueColor24(Tcl_Interp* i, Tk_Canvas c, 
					 Tk_Item* item) : 
  ColorbarTrueColor(i, c, item)
{
  configSpecs = colorbarTrueColor24Specs;  // colorbar configure options

  loadDefaultCMaps();
}

int ColorbarTrueColor24::initColormap()
{
  colorCount = (((ColorbarTrueColor24Options*)options)->colors);
  return initColormapForReal();
}

void ColorbarTrueColor24::updateColors()
{
  updateColorCells();

  // fill in xmap
  // make sure we have a pixmap

  if (!pixmap || !xmap)
    return;

  int& width = options->width;
  int& height = options->height;
  char* data = xmap->data;
    
  // Fill in colorbar data
  // Calculate first row
  // check for bits/pixel

  switch (xmap->bits_per_pixel) {
  case 32:
    updateColors32();
    break;
  case 24:
    updateColors24();
    break;
  default:
    cerr << "ColorbarTrueColor24 Internal Error: bad bits/pixel" << endl;
    exit(1);
  }

  // --and duplicate for remaining rows

  for (int j=1; j<height; j++)
    memcpy(data+(j*xmap->bytes_per_line), data, xmap->bytes_per_line);

  // update the pixmap from map image and clean up

  XPutImage(display, pixmap, gc, xmap, 0, 0, 0, 0, width, height);

  redraw();
}

void ColorbarTrueColor24::updateColors24()
{
  int& width = options->width;
  int& height = options->height;
  char* data = xmap->data;

  int rs = decodeMask((unsigned long)visual->red_mask);
  int gs = decodeMask((unsigned long)visual->green_mask);
  int bs = decodeMask((unsigned long)visual->blue_mask);

  for (int i=0; i<width; i++) {
    unsigned long r = colorCells[(int)(double(i)/width*colorCount)*3+2];
    unsigned long g = colorCells[(int)(double(i)/width*colorCount)*3+1];
    unsigned long b = colorCells[(int)(double(i)/width*colorCount)*3];
    unsigned long result = 0;
    result |= r << rs;
    result |= g << gs;
    result |= b << bs;

    memcpy(data+i*3, &result, 3);
  }
}

void ColorbarTrueColor24::updateColors32()
{
  int& width = options->width;
  int& height = options->height;
  char* data = xmap->data;

  int rs = decodeMask((unsigned long)visual->red_mask);
  int gs = decodeMask((unsigned long)visual->green_mask);
  int bs = decodeMask((unsigned long)visual->blue_mask);

  for (int i=0; i<width; i++) {
    unsigned long r = colorCells[(int)(double(i)/width*colorCount)*3+2];
    unsigned long g = colorCells[(int)(double(i)/width*colorCount)*3+1];
    unsigned long b = colorCells[(int)(double(i)/width*colorCount)*3];
    unsigned long result = 0;
    result |= r << rs;
    result |= g << gs;
    result |= b << bs;

    memcpy(data+i*4, &result, 4);
  }
}
