#include "polygon.h"
#include "frame.h"

Polygon::Polygon(Frame* p, const Vector& ctr, const Vector& b, const char* clr,
		 const char* f, const char* t, unsigned short prop)
  : Marker(p, ctr, clr, f, t, prop)
{
  vertex = new List<Vertex>();

  angle = 0;
  strcpy(type, "polygon");

  Vector bb = b;
  vertex->append(new Vertex(-bb[0],-bb[1]));
  vertex->append(new Vertex( bb[0],-bb[1]));
  vertex->append(new Vertex( bb[0], bb[1]));
  vertex->append(new Vertex(-bb[0], bb[1]));

  updateBBox();
}

Polygon::Polygon(Frame* p, List<Vertex>* list,
		 const char* clr, const char* f, const char* t, 
		 unsigned short prop)
  : Marker(p, Vector(0,0), clr, f, t, prop)
{
  // Vertex list is in ref coords

  angle = 0;
  strcpy(type, "polygon");
  vertex = list;

  // check to see if the first and last node are the same

  if (vertex->head()->vector[0] == vertex->tail()->vector[0] &&
      vertex->head()->vector[1] == vertex->tail()->vector[1])
    delete vertex->pop();

  // find center

  center = Vector(0,0);
  vertex->head();
  do
    center += vertex->current()->vector;
  while (vertex->next());
  center /= vertex->count();

  // vertices are relative

  vertex->head();
  do
    vertex->current()->vector *= Translate(-center); // no rotation
  while (vertex->next());

  updateBBox();
}

Polygon::~Polygon()
{
  if (vertex)
    delete vertex;
}

void Polygon::updateBBox()
{
  // generate handles

  numHandle = 4 + vertex->count();
  if (handle)
    delete [] handle;
  handle = new Vector[numHandle];

  Matrix m = Rotate(-angle) * Translate(center) * parent->refToCanvas;

  // the first four are our control handles

  BBox bb;
  vertex->head();
  do
    bb.bound(vertex->current()->vector);
  while (vertex->next());

  bb.expand(10); // give us more room

  handle[0] = bb.ll   * m;
  handle[1] = bb.lr() * m;
  handle[2] = bb.ur   * m;
  handle[3] = bb.ul() * m;

  // and the rest are vertices

  int i=4;
  vertex->head();
  do
    handle[i++] = vertex->current()->vector * m;
  while (vertex->next());

  // now, calc bbox

  Vector c = center * parent->refToCanvas;
  bbox = BBox(c,c);
  {
    for (int i=0; i<numHandle; i++)
      bbox.bound(handle[i]);
  }
  bbox.expand(3); // give us room for control handles

  // calculate overall bbox

  calcAllBBox();
}

void Polygon::updateCoords(Matrix& m)
{
  Scale s(m);
  vertex->head();
  do
    vertex->current()->vector *= s;
  while (vertex->next());
  
  Marker::updateCoords(m);
}

void Polygon::draw()
{
  Marker::draw();

  includeExclude();
  render(parent->pixmap, parent->refToWidget, gc);
}

void Polygon::drawXOR()
{
  render(Tk_WindowId(parent->tkwin), parent->refToWindow, gcxor);
}

void Polygon::drawMagnifier(const Matrix& m)
{
  XSetForeground(display, gc, color);
  includeExclude();
  render(parent->magnifierPixmap, m, gc);
}

void Polygon::edit(const Vector& v, int h)
{
  if (h < 5) {
    Vector s1 = v * Translate(-center);
    Vector s2 = handle[h-1] * parent->canvasToRef * Translate(-center);

    if (s1[0] != 0 && s1[1] != 0 && s2[0] != 0 && s2[1] != 0) {
      double a = fabs(s1[0]/s2[0]);
      double b = fabs(s1[1]/s2[1]);
      double s = a > b ? a : b;

      vertex->head();
      do
	vertex->current()->vector *= Scale(s);
      while (vertex->next());
    }

    updateBBox();
    doCallBacks(&editCB);
  }
  else {
    moveVertex(v,h);

    updateBBox();
    doCallBacks(&editCB);
    doCallBacks(&moveCB); // center can change
  }
}

void Polygon::rotate(const Vector& v, int h)
{
  if (h < 5) {
    // v is in ref coords
    // handles are in canvas coords

    Vector a = v * Translate(-center);
    Vector b = handle[h-1] * parent->canvasToRef * Translate(-center);

    angle += a.angle() - b.angle();

    updateBBox();
    doCallBacks(&rotateCB);
  }
  else {
    // we need to check this here, because we are really rotating
    if (canEdit()) { 
      moveVertex(v,h);

      updateBBox();
      doCallBacks(&editCB);
      doCallBacks(&moveCB); // center can change
    }
  }
}

void Polygon::ps(int mode)
{
  Marker::ps(mode);
  psIncludeExclude();

  Matrix m = Rotate(-angle) * Translate(center) * parent->refToCanvas;

  char buf[256];
  vertex->head();
  int first = 1;
  do {
    ostrstream str(buf,256);
    Vector v =  vertex->current()->vector * m;
    if (first) {
      str << "newpath " << endl
	  << v.TkCanvasPs(parent->canvas) << " moveto" << endl << ends;
      first = 0;
    }
    else
      str << v.TkCanvasPs(parent->canvas) << " lineto" << endl << ends;

    Tcl_AppendResult(parent->interp, buf, NULL);
  } while (vertex->next());

  ostrstream str(buf,256);
  str << "closepath stroke" << endl << ends;
  Tcl_AppendResult(parent->interp, buf, NULL);
}

int Polygon::isIn(const Vector& vv)
{
  /*
    v[0]-- x value of point being tested
    v[1]-- y value of point being tested

    This algorithm is from "An Introduction to Ray Tracing", Academic Press,
    1989, edited by Andrew Glassner, pg 53
    -- a point lies in a polygon if a line is extended from the point to 
    infinite in any direction and the number of intersections with the 
    polygon is odd.
    This is valid for both concave and convex polygons.
    Points on a vertex are considered inside.
    Points on a edge are considered inside.
  */

  Vector v = vv * parent->canvasToRef; // xform to ref coords
  int crossings = 0;   // number of crossings

  vertex->head();
  Vector v1;
  Vector v2 = (vertex->current()->vector * 
	       Rotate(-angle) * 
	       Translate(center)) - v;

  int sign = ((v2[1])>=0) ? 1 : -1; // init sign

  // for all edges

  int done = 0;

  do {
    // look at next two vertices
    v1 = v2;
    if (!vertex->next()) {
      done = 1;
      vertex->head();
    }
    v2 = (vertex->current()->vector * Rotate(-angle) * Translate(center)) - v;

    int nextSign = (v2[1]>=0) ? 1 : -1; // sign holder for p2

    if (sign != nextSign) {
      if (v1[0]>0 && v2[0]>0)
	crossings++;
      else if (v1[0]>0 || v2[0]>0) {
	if (v1[0]-(v1[1]*(v2[0]-v1[0])/(v2[1]-v1[1])) > 0)
	  crossings++;
      }
      sign = nextSign;
    }
  } while (!done);

  return fmod(crossings,2) ? 1 : 0; // if odd, point is inside
}

void Polygon::setAngle(double a)
{
  angle = a;
  updateBBox();

  doCallBacks(&rotateCB);
}

void Polygon::createVertex(int which, const Vector& v)
{
  // which segment (1 to n)
  // v is in ref coords

  int seg = which-1;
  if (seg>=0 && seg<vertex->count()) {
    Vertex* n = new Vertex(v * Translate(-center) * Rotate(angle));
    vertex->insert(seg,n);

    updateBBox();
    doCallBacks(&editCB);
    doCallBacks(&moveCB); // center can change
  }
}

void Polygon::deleteVertex(int h)
{
  if (vertex->count() > 3) {
    Vertex* v = (*vertex)[h];
    if (v) {
      vertex->extractNext(v);

      updateBBox();
      doCallBacks(&editCB);
      doCallBacks(&moveCB); // center can change
    }  
  }
}

int Polygon::getSegment(const Vector& v)
{
  // v is in canvas coords

  Matrix m = Rotate(angle) * Translate(center);
  vertex->head();
  Vector v1;
  Vector v2 = vertex->current()->vector * m;
  int done = 0;
  int i = 1;
  do {
    v1 = v2;
    if (!vertex->next()) {
      vertex->head();
      done = 1;
    }
    v2 = vertex->current()->vector * m;

    Vector l1 = v1*parent->refToCanvas;
    Vector l2 = v2*parent->refToCanvas;
    double a = (l2-l1).angle();
    Matrix mm = Translate(-l1) * Rotate(a); 
    Vector end = l2*mm;
    Vector vv = v*mm;
    
    if (vv[0]>0 && vv[0]<end[0] && vv[1]>-3 && vv[1]<3)
      return i;

    i++;
  } while (!done);

  return 0;
}

// private

void Polygon::render(Drawable drawable, const Matrix& matrix, GC ggc)
{
  Matrix m = Rotate(-angle) * Translate(center) * matrix;

  vertex->head();
  Vector v1;
  Vector v2 = (vertex->current()->vector * m).round();
  int done = 0;

  do {
    if (!vertex->next()) {
      done = 1;
      vertex->head();
    }
    v1 = v2;
    v2 = (vertex->current()->vector * m).round();
    XDrawLine(display, drawable, ggc, (int)v1[0], (int)v1[1], 
	      (int)v2[0], (int)v2[1]);
  } while (!done);
}

void Polygon::moveVertex(const Vector& v, int h)
{
  if ((*vertex)[h-5])
    vertex->current()->vector = v * Translate(-center) * Rotate(angle);

  // recalculate center

  Vector nc;
  vertex->head();
  do
    nc += vertex->current()->vector * Rotate(-angle);
  while (vertex->next());
  nc /= vertex->count();

  center += nc;

  // update all vertices

  vertex->head();
  do
    vertex->current()->vector -= nc * Rotate(angle);
  while (vertex->next());
}

// list

void Polygon::list(ostream& str, int sys, int format)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::PHYSICAL:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      str << type;
      vertex->head();
      do {
	Vector w = vertex->current()->vector * 
	  Rotate(-angle) * 
	  Translate(center);
	Vector v = parent->map(w,Frame::REF,(Frame::CoordSystem)sys);
	str << setprecision(8) << v;
      }
      while (vertex->next());
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::ALTAZ:
  case Frame::LINEAR:
  case Frame::ICRS:
    str << type;
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	vertex->head();
	do {
	  Vector w = vertex->current()->vector * 
	    Rotate(-angle) * 
	    Translate(center);
	  Vector v = parent->map(w,Frame::REF,(Frame::CoordSystem)sys);
	  str << setprecision(8) << v;
	}
	while (vertex->next());
      }
      break;
    case Frame::HMS:
      {
	char buf[64];
	char ra[16];
	char dec[16];

	vertex->head();
	do {
	  Vector w = vertex->current()->vector * 
	    Rotate(-angle) * 
	    Translate(center);
	  parent->mapToWCSStr(w, Frame::REF, (Frame::CoordSystem)sys,
			      (Frame::CoordFormat)format, buf, 64);
	  istrstream wcs(buf,64);
	  wcs >> ra >> dec;
	  str << ' ' << ra << ' ' << dec;
	}
	while (vertex->next());
      }
      break;
    }
    break;
  }

  listProperties(str);
}

void Polygon::listSAOtng(ostream& str, int sys, int format)
{
  if (text[0])
    str << '#' << text << endl;

  if (properties&INCLUDE)
    str << '+';
  else
    str << '-';

  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::PHYSICAL:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Frame::CoordSystem toSys = (Frame::CoordSystem)sys;
      if (toSys == Frame::DETECTOR || toSys == Frame::AMPLIFIER)
	toSys = Frame::IMAGE;

      str << type << '(';
      int first=1;
      vertex->head();
      do {
	if (!first)
	  str << ',';
	first=0;

	Vector w = vertex->current()->vector * 
	  Rotate(-angle) * 
	  Translate(center);
	Vector v = parent->map(w,Frame::REF,toSys);
	str << setprecision(8) << v[0] << ',' << v[1];
      }
      while (vertex->next());
      str << ") ";
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::ALTAZ:
  case Frame::LINEAR:
  case Frame::ICRS:
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	str << type << '(';
	int first=1;
	vertex->head();
	do {
	  if (!first)
	    str << ',';
	  first=0;

	  Vector w = vertex->current()->vector * 
	    Rotate(-angle) * 
	    Translate(center);
	  Vector v = parent->map(w,Frame::REF,(Frame::CoordSystem)sys);
	  str << setprecision(8) << v[0] << ',' << v[1];
	}
	while (vertex->next());
	str << ") ";
      }
      break;
    case Frame::HMS:
      {
	char buf[64];
	char ra[16];
	char dec[16];

	str << type << '(';
	int first=1;
	vertex->head();
	do {
	  if (!first)
	    str << ',';
	  first=0;

	  Vector w = vertex->current()->vector * 
	    Rotate(-angle) * 
	    Translate(center);
	  parent->mapToWCSStr(w, Frame::REF, (Frame::CoordSystem)sys,
			      (Frame::CoordFormat)format, buf, 64);
	  istrstream wcs(buf,64);
	  wcs >> ra >> dec;
	  str << ra << ',' << dec;
	}
	while (vertex->next());
	str << ") ";
      }
      break;
    }
    break;
  }

  str << "# " << colorName;
}

void Polygon::listPros(ostream& str, int sys, int format)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      str << "polygon ";
      vertex->head();
      do {
	Vector w = vertex->current()->vector * 
	  Rotate(-angle) * 
	  Translate(center);
	Vector v = parent->map(w,Frame::REF,Frame::IMAGE);
	str << setprecision(8) << v;
      }
      while (vertex->next());
    }
    break;

  case Frame::PHYSICAL:
    {
      str << "physical; polygon ";
      vertex->head();
      do {
	Vector w = vertex->current()->vector * 
	  Rotate(-angle) * 
	  Translate(center);
	Vector v = parent->map(w,Frame::REF,(Frame::CoordSystem)sys);
	str << setprecision(8) << v;
      }
      while (vertex->next());
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::ALTAZ:
  case Frame::LINEAR:
  case Frame::ICRS:
    parent->listProsCoordSystem(str, (Frame::CoordSystem)sys);
    str << "; polygon ";
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	vertex->head();
	do {
	  Vector w = vertex->current()->vector * 
	    Rotate(-angle) * 
	    Translate(center);
	  Vector v = parent->map(w,Frame::REF,(Frame::CoordSystem)sys);
	  str << setprecision(8) << v[0] << "d " << v[1] << "d ";
	}
	while (vertex->next());
      }
      break;
    case Frame::HMS:
      {
	char buf[64];
	char ra[16];
	char dec[16];

	vertex->head();
	do {
	  Vector w = vertex->current()->vector *
	    Rotate(-angle) * 
	    Translate(center);
	  parent->mapToWCSStr(w, Frame::REF, (Frame::CoordSystem)sys,
			      (Frame::CoordFormat)format, buf, 64);
	  istrstream wcs(buf,64);
	  wcs >> ra >> dec;
	  if (dec[0]=='+')
	    str << ' ' << ra << ' ' << dec+1;
	  else
	    str << ' ' << ra << ' ' << dec;
	}
	while (vertex->next());
      }
      break;
    }
    break;
  }
}

void Polygon::listSAOimage(ostream& str)
{
  if (!(properties&INCLUDE))
    str << '-';

  str << "polygon(";
  int first=1;
  vertex->head();
  do {
    if (!first)
      str << ',';
    first=0;

    Vector w = vertex->current()->vector * Rotate(-angle) * Translate(center);
    Vector v = parent->map(w,Frame::REF,Frame::IMAGE);
    str << setprecision(8) << v[0] << ',' << v[1];
  }
  while (vertex->next());
  str << ')';
}
