#include "ruler.h"
#include "frame.h"
#include "util.h"

Ruler::Ruler(Frame* p, const Vector& ctr, const Vector& ptr, int format,
	     const char* clr, const char* f, const char* t, 
	     unsigned short prop) 
  : Marker(p, ctr, clr, f, t, prop)
{
  coordSystem = Frame::IMAGE;
  coordFormat = format;
  dist = 0;
  gcxor = parent->markerGCXORRuler;

  p2 = ptr;
  strcpy(type,"ruler");
  handle = new Vector[2];
  numHandle = 2;

  updateBBox();
}

void Ruler::updateBBox()
{
  calcP3();
  calcDist();
  calcAxisDist();

  // bound marker

  bbox = BBox(center * parent->refToCanvas, p2 * parent->refToCanvas);
  bbox.bound(p3 * parent->refToCanvas);

  // generate handles in canvas coords

  handle[0] = center * parent->refToCanvas;
  handle[1] = p2 * parent->refToCanvas;

  // make room for handles/arrows/text

  bbox.expand(40);

  // calculate overall bbox

  calcAllBBox();
}

void Ruler::updateCoords(Matrix& m)
{
  p2*=m;
  p3*=m;
  Marker::updateCoords(m);
}

void Ruler::draw()
{
  XSetForeground(display, gc, color);
  drawHandles();
  drawText();
  
  render(parent->pixmap, parent->refToWidget, gc);
}

void Ruler::drawXOR()
{
  XSetForeground(display, gcxor, parent->getColor("white"));
  render(Tk_WindowId(parent->tkwin), parent->refToWindow, gcxor);
}

void Ruler::drawMagnifier(const Matrix& m)
{
  XSetForeground(display, gc, color);
  render(parent->magnifierPixmap, m, gc);
}

void Ruler::move(const Vector& v)
{
  center+=v;
  p2+=v;

  updateBBox();
  doCallBacks(&moveCB);
}

void Ruler::moveTo(const Vector& v)
{
  Vector diff = v - center;
  center=v;
  p2+=diff;

  updateBBox();
  doCallBacks(&moveCB);
}

void Ruler::edit(const Vector& v, int h)
{
  switch (h) {
  case 1:
    center = v;
    break;
  case 2:
    p2 = v;
    break;
  }

  updateBBox();
  doCallBacks(&editCB);
}

void Ruler::ps(int mode)
{
  Vector a = center * parent->refToCanvas;
  Vector b = p2 * parent->refToCanvas;
  Vector c = p3 * parent->refToCanvas;

  Marker::ps(mode);

  char buf[1024];
  ostrstream str(buf,1024);
  str << "[] 0 setdash" << endl;

  str << "newpath " 
      << a.TkCanvasPs(parent->canvas) << "moveto"
      << b.TkCanvasPs(parent->canvas) << "lineto"
      << " stroke" << endl;

  str << "[8 3] 0 setdash" << endl;

  str << "newpath " 
      << a.TkCanvasPs(parent->canvas) << "moveto"
      << c.TkCanvasPs(parent->canvas) << "lineto"
      << " stroke" << endl;

  str << "[8 3] 0 setdash" << endl;

  str << "newpath " 
      << b.TkCanvasPs(parent->canvas) << "moveto"
      << c.TkCanvasPs(parent->canvas) << "lineto"
      << " stroke" << endl;

  str << ends;

  Tcl_AppendResult(parent->interp, buf, NULL);

  psArrow(a, (b-a).normalize());
  psArrow(b, (a-b).normalize());

  psValue(mode, (b-a)/2 + a, dist);
}

int Ruler::isOn(const Vector& v, const Vector& v1, const Vector& v2)
{
  // v : canvas coords
  // v1: ref coords
  // v2: ref coords

  // do this in canvas coords, not ref coords

  Vector l1 = v1 * parent->refToCanvas;
  Vector l2 = v2 * parent->refToCanvas;
  double a = (l2-l1).angle();

  Matrix m = Translate(-l1) * Rotate(a);
  Vector end = l2*m;
  Vector vv = v*m;
  return (vv[0]>0 && vv[0]<end[0] && vv[1]>-3 && vv[1]<3);
}

int Ruler::isIn(const Vector& vv)
{
  // test to see if point is on edge, if so, considered inside

  if (isOn(vv,center,p2) || isOn(vv,center,p3) || isOn(vv,p2,p3))
    return 1;

  /*
    v[0]-- x value of point being tested
    v[1]-- y value of point being tested

    This algorithm is from "An Introduction to Ray Tracing", Academic Press,
    1989, edited by Andrew Glassner, pg 53
    -- a point lies in a polygon if a line is extended from the point to 
    infinite
    in any direction and the number of intersections with the polygon is odd.
    This is valid for both concave and convex polygons.
    Points on a vertex are considered inside.
    Points on a edge are considered inside.
  */

  // analysis in ref coords
  Vector v = vv * parent->canvasToRef;

  int crossings = 0;   // number of crossings
  int sign;

  // for all edges

  for (int i=0; i<3; i++) {
    Vector v1, v2;
    switch (i) {
    case 0:
      v1 = center-v;
      v2 = p2-v;
      sign = ((v1[1])>=0) ? 1 : -1; // init sign
      break;
    case 1:
      v1 = p2-v;
      v2 = p3-v;
      break;
    case 2:
      v1 = p3-v;
      v2 = center-v;
      break;
    }

    int nextSign = (v2[1]>=0) ? 1 : -1; // sign holder for p2

    if (sign != nextSign) {
      if (v1[0]>0 && v2[0]>0)
	crossings++;
      else if (v1[0]>0 || v2[0]>0) {
	if (v1[0]-(v1[1]*(v2[0]-v1[0])/(v2[1]-v1[1])) > 0)
	  crossings++;
      }
      sign = nextSign;
    }
  }

  return fmod(crossings,2) ? 1 : 0; // if odd, point is inside
}

void Ruler::setP1(const Vector& v)
{
  center = v;

  updateBBox();
  doCallBacks(&editCB);
}

void Ruler::setP2(const Vector& v)
{
  p2 = v;

  updateBBox();
  doCallBacks(&editCB);
}

void Ruler::setSystem(int sys, int format)
{
  coordSystem = sys;
  coordFormat = format;

  updateBBox();
}

// Private

void Ruler::render(Drawable drawable, const Matrix& matrix, GC ggc)
{
  Vector a = (center * matrix).round();
  Vector b = (p2 * matrix).round();
  Vector c = (p3 * matrix).round();

  noDash(ggc);
  XDrawLine(display, drawable, ggc, (int)a[0], (int)a[1], 
	    (int)b[0], (int)b[1]);
  renderArrow(drawable, a, (b-a).normalize(), ggc);
  renderArrow(drawable, b, (a-b).normalize(), ggc);

  dash(ggc);
  XDrawLine(display, drawable, ggc, (int)a[0], (int)a[1], 
	    (int)c[0], (int)c[1]);
  XDrawLine(display, drawable, ggc, (int)b[0], (int)b[1], 
	    (int)c[0], (int)c[1]);

  renderValue(drawable, (b-a)/2 + a, dist, ggc);
}

void Ruler::renderValue(Drawable drawable, Vector v, double value, GC ggc)
{
  Tk_Font df = Tk_GetFont(parent->interp, parent->tkwin, "system");
  Tk_FontMetrics metrics;
  Tk_GetFontMetrics(df, &metrics);

  char buf[256];
  ostrstream str(buf,256);
  switch ((Frame::CoordFormat)coordFormat) {
    case Frame::PIXELS:
      str << value << ends;
      break;
    case Frame::DEGREES:
      str << value << " deg" << ends;
      break;
    case Frame::ARCMIN:
      str << value << '\'' << ends;
      break;
    case Frame::ARCSEC:
      str << value << '"' << ends;
      break;
  }

  int w = Tk_TextWidth(df, buf, strlen(buf))+strlen(buf);
  int h = metrics.descent + metrics.ascent+2;
  int x = (int)(v[0]-w/2.);
  int y = (int)v[1];
  XSetForeground(display, ggc, parent->getColor("yellow"));
  XFillRectangle(display, drawable, ggc, x, y-metrics.ascent, w, h);

  XSetForeground(display, ggc, parent->getColor("black"));
  XSetFont(display, ggc, Tk_FontId(df));
  XDrawString(display, drawable, ggc, x+1, y+1, buf, strlen(buf));

  Tk_FreeFont(df);
}

void Ruler::psValue(int mode, Vector v, double value)
{
  // create value string

  char vbuf[32];
  ostrstream vstr(vbuf,32);
  switch ((Frame::CoordFormat)coordFormat) {
  case Frame::PIXELS:
    vstr << value << ends;
    break;
  case Frame::DEGREES:
    vstr << value << " deg" << ends;
    break;
  case Frame::ARCMIN:
    vstr << value << '\'' << ends;
    break;
  case Frame::ARCSEC:
    vstr << value << '"' << ends;
    break;
  }

  // define font

  char fontstr[] = "system 12 normal";
  char buf[512];
  ostrstream str(buf,512);

  Tcl_DString psFont;
  Tcl_DStringInit(&psFont);
  Tk_Font df = Tk_GetFont(parent->interp, parent->tkwin, fontstr);
  int psSize = 
    Tk_PostscriptFontName(df, &psFont);

  Tk_FontMetrics metrics;
  Tk_GetFontMetrics(df, &metrics);

  str << '/' << Tcl_DStringValue(&psFont) 
      << " findfont " << psSize << " scalefont setfont" << endl;

  Tcl_DStringFree(&psFont);

  // do rect

  switch (mode) {
  case 1:
    psColorGray("yellow", str);
    str << " setgray";
    break;
  case 2:
    psColorRGB("yellow", str);
    str << " setrgbcolor";
    break;
  }
  str << endl;

  int width = Tk_TextWidth(df, vbuf, strlen(vbuf))-strlen(vbuf); // it works...
  int height = metrics.ascent;
  str << (v-Vector(width, 0)/2).TkCanvasPs(parent->canvas) << ' '
      << width << ' ' << height << " rectfill" << endl;

  // do text

  switch (mode) {
  case 1:
    psColorGray("black", str);
    str << " setgray";
    break;
  case 2:
    psColorRGB("black", str);
    str << " setrgbcolor";
    break;
  }
  str << endl;

  str << v.TkCanvasPs(parent->canvas) << " moveto" << endl
      << '(' << vbuf << ')' << endl
      << "dup stringwidth pop 2 div 0 exch sub " 
      << metrics.descent << " rmoveto show" << endl;

  // done

  str << ends;
  Tcl_AppendResult(parent->interp, buf, NULL);

  Tk_FreeFont(df);
}

void Ruler::calcP3()
{
  Vector a = parent->map(center, Frame::REF, (Frame::CoordSystem)coordSystem);
  Vector b = parent->map(p2, Frame::REF, (Frame::CoordSystem)coordSystem);
  p3 = parent->map(Vector(b[0],a[1]), (Frame::CoordSystem)coordSystem, Frame::REF);
}

void Ruler::calcDist()
{
  dist = parent->mapDistFromRef(p2, center, 
				(Frame::CoordSystem)coordSystem,
				(Frame::CoordFormat)coordFormat);
}

void Ruler::calcAxisDist()
{
  double x = parent->mapDistFromRef(p3, center, 
				    (Frame::CoordSystem)coordSystem,
				    (Frame::CoordFormat)coordFormat);
  double y = parent->mapDistFromRef(p3, p2, 
				    (Frame::CoordSystem)coordSystem,
				    (Frame::CoordFormat)coordFormat);
  axisDist = Vector(x,y);
}

// list

void Ruler::list(ostream& str, int sys, int format)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::PHYSICAL:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v1 = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      Vector v2 = parent->map(p2,Frame::REF,(Frame::CoordSystem)sys);
      str << type << setprecision(8) << v1 << v2;
      parent->listCoordFormat(str, (Frame::CoordFormat)coordFormat);
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::ALTAZ:
  case Frame::LINEAR:
  case Frame::ICRS:
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v1 = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	Vector v2 = parent->map(p2,Frame::REF,(Frame::CoordSystem)sys);
	str << type << setprecision(8) << v1 << v2;
	parent->listCoordFormat(str, (Frame::CoordFormat)coordFormat);
      }
      break;
    case Frame::HMS:
      {
	char buf[64];
	char ra1[16], ra2[16];
	char dec1[16], dec2[16];
	{
	  parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			      (Frame::CoordFormat)format, buf, 64);
	  istrstream wcs(buf,64);
	  wcs >> ra1 >> dec1;
	}
	{
	  parent->mapToWCSStr(p2, Frame::REF, (Frame::CoordSystem)sys,
			      (Frame::CoordFormat)format, buf, 64);
	  istrstream wcs(buf,64);
	  wcs >> ra2 >> dec2;
	}
	str << type << ' ' << ra1 << ' ' << dec1 << ' ' << ra2 << ' ' << dec2 << ' ';
	parent->listCoordFormat(str, (Frame::CoordFormat)coordFormat);
      }
      break;
    }
    break;
  }

  listProperties(str);
}
